import datetime

import pytest

from valorant import MISSING, AbilityType, Agent, Event, RelationType, RewardType, Season

from .conftest import BaseTest


class TestValorantAPI(BaseTest):
    @pytest.mark.asyncio
    async def test_fetch_assets(self) -> None:
        await self.client.init()

    @pytest.mark.asyncio
    async def test_agents(self) -> None:
        assert len(self.client.agents) > 0
        for agent in self.client.agents:
            # fagent = await self.client.fetch_agent(agent.uuid)
            fagent = self.client.get_agent(agent.uuid)
            assert fagent is not None
            assert fagent == agent

            assert agent.display_name is not None
            assert agent.description is not None
            assert agent.display_icon is not None
            assert agent.bust_portrait is not None
            assert agent.full_portrait is not None
            assert agent.full_portrait_v2 is not None
            assert agent.killfeed_portrait is not None
            assert agent.background is not None

            assert agent.role is not None
            assert agent.role.display_name is not None
            assert agent.role.description is not None
            assert agent.role.display_icon is not None

            assert agent.abilities is not None
            assert isinstance(agent.abilities, list)
            assert len(agent.abilities) >= 4 or len(agent.abilities) <= 5
            for ability in agent.abilities:
                assert ability is not None
                assert ability.slot is not None
                assert isinstance(ability.slot, AbilityType)
                assert ability.display_name is not None
                assert ability.description is not None
                if ability.display_icon:
                    assert ability.display_icon is not None

            assert agent.voice_line_localization is not None
            voice_line = agent.voice_line
            if voice_line is not None:
                assert agent.voice_line_localization.voice_locale == voice_line
                for voice in agent.voice_line_localization.all():
                    assert voice is not None
                    assert voice.min_duration is not None
                    assert voice.max_duration is not None
                    assert voice.media_list is not None
                    for media in voice.media_list:
                        assert media is not None
                        assert media.id is not None
                        assert media.wwise is not None
                        assert media.wave is not None

    @pytest.mark.asyncio
    async def test_buddies(self) -> None:
        assert len(self.client.buddies) > 0
        for buddy in self.client.buddies:
            assert buddy is not None
            # fbuddy = await self.client.fetch_buddy(buddy.uuid)
            fbuddy = self.client.get_buddy(buddy.uuid)
            assert fbuddy is not None
            assert fbuddy == buddy
            assert buddy is not None
            assert buddy.display_name is not None
            assert isinstance(buddy.is_hidden_if_not_owned(), bool)
            if buddy._theme_uuid is not None:
                assert buddy.theme is not None
            assert buddy.display_icon is not None
            assert buddy.asset_path is not None
            assert len(buddy.levels) > 0
            assert buddy.get_buddy_level(1) is not None

            for level in buddy.levels:
                assert level is not None
                assert level.parent is not None
                assert level.charm_level is not None
                assert level.display_name is not None
                assert level.display_icon is not None
                assert level.asset_path is not None

    @pytest.mark.asyncio
    async def test_bundles(self) -> None:
        assert len(self.client.bundles) > 0
        for bundle in self.client.bundles:
            # fbundle = await self.client.fetch_bundle(bundle.uuid)
            fbundle = self.client.get_bundle(bundle.uuid)
            assert fbundle is not None
            assert fbundle == bundle

            assert bundle is not None
            assert bundle.display_name is not None
            assert bundle.description is not None
            if bundle.display_name_sub_text:
                assert bundle.display_name_sub_text is not None
            if bundle.extra_description:
                assert bundle.extra_description is not None
            if bundle.promo_description:
                assert bundle.promo_description is not None
            assert isinstance(bundle.use_additional_context, bool)
            assert bundle.display_icon is not None
            assert bundle.display_icon_2 is not None
            if bundle.vertical_promo_image:
                assert bundle.vertical_promo_image is not None
            assert bundle.asset_path is not None

    @pytest.mark.asyncio
    async def test_ceremonies(self) -> None:
        assert len(self.client.ceremonies) > 0
        for ceremony in self.client.ceremonies:
            assert ceremony is not None
            assert ceremony.display_name is not None
            assert ceremony.asset_path is not None

    @pytest.mark.asyncio
    async def test_competitive_tiers(self) -> None:
        assert len(self.client.competitive_tiers) > 0
        for c_tier in self.client.competitive_tiers:
            assert c_tier is not None
            assert c_tier.asset_path is not None
            assert len(c_tier.tiers) > 0
            for tier in c_tier.tiers:
                assert tier is not None
                assert tier.tier is not None
                assert tier.name is not None
                assert tier.division is not None
                assert tier.division_name is not None
                assert tier.color is not None
                assert tier.background_color is not None
                if tier.small_icon:
                    assert tier.small_icon is not None
                if tier.large_icon:
                    assert tier.large_icon is not None
                if tier.rank_triangle_down_icon:
                    assert tier.rank_triangle_down_icon is not None
                if tier.rank_triangle_up_icon:
                    assert tier.rank_triangle_up_icon is not None

    @pytest.mark.asyncio
    async def test_content_tiers(self) -> None:
        assert len(self.client.content_tiers) > 0
        for content_tier in self.client.content_tiers:
            assert content_tier is not None
            assert content_tier.display_name is not None
            assert content_tier.dev_name is not None
            assert content_tier.rank is not None
            assert content_tier.juice_value is not None
            assert content_tier.juice_cost is not None
            assert content_tier.highlight_color is not None
            assert content_tier.display_icon is not None
            assert content_tier.asset_path is not None

    @pytest.mark.asyncio
    async def test_contracts(self) -> None:
        assert len(self.client.contracts) > 0
        for contract in self.client.contracts:
            assert contract is not None
            assert contract.display_name is not None
            if contract.display_icon:
                assert contract.display_icon is not None
            assert isinstance(contract.ship_it, bool)
            assert contract.free_reward_schedule_uuid is not None
            assert contract.asset_path is not None
            content = contract.content
            assert content is not None
            assert isinstance(content.relation_type, RelationType)
            assert content.premium_vp_cost is not None
            # assert content.premium_reward_schedule_uuid is not None
            if content._relation_uuid is None:
                assert content.relationship is None
            if content.relationship is not None:
                assert isinstance(content.relationship, (Agent, Event, Season))
            assert len(content.chapters) > 0
            for chapter in content.chapters:
                assert chapter is not None
                assert isinstance(chapter.is_epilogue(), bool)
                assert len(chapter.levels) > 0
                for level in chapter.levels:
                    assert level is not None
                    assert level.reward is not None
                    reward = level.reward
                    assert reward.type is not None
                    assert isinstance(reward.type, RewardType)
                    assert reward.amount is not None
                    assert isinstance(reward.amount, int)
                    item = reward.get_item()
                    assert item is not None  # if item is None, maybe game is updated new reward type
                    assert isinstance(level.xp, int)
                    assert isinstance(level.vp_cost, int)
                    assert isinstance(level.dough_cost, int)
                    assert isinstance(level.is_purchasable_with_vp(), bool)
                    assert isinstance(level.is_purchasable_with_dough(), bool)

                if chapter.free_rewards is not None:
                    assert len(chapter.free_rewards) > 0
                    for reward in chapter.free_rewards:
                        assert reward is not None
                        assert reward.type is not None
                        assert isinstance(reward.type, RewardType)
                        assert reward.amount is not None
                        assert isinstance(reward.amount, int)
                        item_free = reward.get_item()
                        assert item_free is not None  # if item is None, maybe game is updated new reward type

    @pytest.mark.asyncio
    async def test_currencies(self) -> None:
        assert len(self.client.currencies) > 0
        for currency in self.client.currencies:
            assert currency is not None
            assert currency.display_name is not None
            assert currency.display_name_singular is not None
            assert currency.display_icon is not None
            assert currency.large_icon is not None
            assert currency.asset_path is not None

    @pytest.mark.asyncio
    async def test_events(self) -> None:
        assert len(self.client.events) > 0
        for event in self.client.events:
            assert event is not None
            assert event.display_name is not None
            assert event.short_display_name is not None
            assert event.start_time is not None
            assert isinstance(event.start_time, datetime.datetime)
            assert event.end_time is not None
            assert isinstance(event.end_time, datetime.datetime)
            assert event.asset_path is not None

    @pytest.mark.asyncio
    async def test_gamemodes(self) -> None:
        assert len(self.client.game_modes) > 0
        for gm in self.client.game_modes:
            assert gm is not None
            assert gm.display_name is not None
            assert isinstance(gm.allows_match_timeouts, bool)
            assert isinstance(gm.is_team_voice_allowed(), bool)
            assert isinstance(gm.is_minimap_hidden(), bool)
            assert gm.orb_count is not None
            assert isinstance(gm.orb_count, int)
            assert gm.rounds_per_half is not None
            assert isinstance(gm.rounds_per_half, int)
            team_roles = gm.team_roles
            if team_roles is not None:
                assert isinstance(team_roles, list)
                assert len(team_roles) > 0
                for tr in team_roles:
                    assert tr is not None
                    assert isinstance(tr, str)
            feature_overrides = gm.game_feature_overrides
            if feature_overrides is not None:
                assert len(feature_overrides) > 0
                for fo in feature_overrides:
                    assert fo is not None
                    assert fo.feature_name is not None
                    assert isinstance(fo.state, bool)
            game_rule_bool_overrides = gm.game_rule_bool_overrides
            if game_rule_bool_overrides is not None:
                assert len(game_rule_bool_overrides) > 0
                for gro in game_rule_bool_overrides:
                    assert gro is not None
                    assert gro.rule_name is not None
                    assert isinstance(gro.state, bool)
            if gm.display_icon:
                assert gm.display_icon is not None
            assert gm.asset_path is not None

    @pytest.mark.asyncio
    async def test_gamemode_equippables(self) -> None:
        assert len(self.client.game_mode_equippables) > 0
        for gme in self.client.game_mode_equippables:
            assert gme is not None
            assert gme.display_name is not None
            assert gme.display_icon is not None
            assert gme.kill_stream_icon is not None
            assert gme.asset_path is not None
            # weapon = gme.get_weapon()

    @pytest.mark.asyncio
    async def test_gear(self) -> None:
        assert len(self.client.gear) > 0
        for gear in self.client.gear:
            assert gear is not None
            assert gear.display_name is not None
            assert gear.display_icon is not None
            shop_data = gear.shop_data
            assert shop_data is not None
            assert shop_data.cost is not None
            assert shop_data.category is not None
            assert shop_data.category_text is not None
            grid_position = shop_data.grid_position
            if grid_position is not None:
                assert grid_position.row is not None
                assert grid_position.column is not None
            assert shop_data.can_be_trashed is not None
            assert isinstance(shop_data.can_be_trashed, bool)
            if shop_data.new_image:
                assert shop_data.new_image is not None
            assert shop_data.new_image is not None
            if shop_data.new_image_2:
                assert shop_data.new_image_2 is not None
            assert shop_data.asset_path is not None
            assert gear.asset_path is not None

    @pytest.mark.asyncio
    async def test_level_borders(self) -> None:
        assert len(self.client.level_borders) > 0
        for border in self.client.level_borders:
            assert border is not None
            assert border.starting_level is not None
            assert border.level_number_appearance is not None
            assert border.small_player_card_appearance is not None
            assert border.asset_path is not None

    @pytest.mark.asyncio
    async def test_maps(self) -> None:
        assert len(self.client.maps) > 0
        for mp in self.client.maps:
            assert mp is not None
            assert mp.display_name is not None
            assert mp.coordinates is not None
            assert mp.display_name is not None
            assert mp.list_view_icon is not None
            assert mp.asset_path is not None
            assert mp.url is not None
            assert mp.x_multiplier is not None
            assert mp.y_multiplier is not None
            assert mp.x_scalar_to_add is not None
            assert mp.y_scalar_to_add is not None
            callouts = mp.callouts
            if callouts is not None:
                for callout in callouts:
                    assert callout.region_name is not None
                    assert callout.super_region_name is not None
                    assert callout.location is not None
                    assert callout.location.x is not None
                    assert callout.location.y is not None

    @pytest.mark.asyncio
    async def test_missions(self) -> None:
        assert len(self.client.missions) > 0
        for mission in self.client.missions:
            assert mission is not None
            if mission.display_name:
                assert mission.display_name is not None
            if mission.title:
                assert mission.title is not None
            if mission.type:
                assert mission.type is not None
            assert mission.xp_grant is not None
            assert mission.progress_to_complete is not None
            assert mission.activation_date is not None
            assert mission.expiration_date is not None
            if mission.tags is not None:
                assert len(mission.tags) > 0
            if mission.objectives:
                assert len(mission.objectives) > 0
                for objective in mission.objectives:
                    assert objective is not None
                    assert objective.value is not None
                    assert isinstance(objective.value, int)
            assert mission.asset_path is not None

    @pytest.mark.asyncio
    async def test_player_cards(self) -> None:
        assert len(self.client.player_cards) > 0
        for card in self.client.player_cards:
            assert card is not None
            assert card.display_name is not None
            assert card.display_icon is not None
            assert card.asset_path is not None
            assert isinstance(card.is_hidden_if_not_owned(), bool)
            assert card.small_art is not None
            assert card.wide_art is not None
            assert card.large_art is not None
            if card._theme_uuid is not None:
                assert card.theme is not None

    @pytest.mark.asyncio
    async def test_player_titles(self) -> None:
        assert len(self.client.player_titles) > 0
        for title in self.client.player_titles:
            assert title is not None
            assert title.display_name is not None
            assert title.text is not None
            assert isinstance(title.is_hidden_if_not_owned(), bool)
            assert title.asset_path is not None

    @pytest.mark.asyncio
    async def test_seasons(self) -> None:
        assert len(self.client.seasons) > 0
        for season in self.client.seasons:
            assert season is not None
            assert season.display_name is not None
            if season.type is not None:
                assert season.type.lower() == 'act'
            assert season.start_time is not None
            assert season.end_time is not None
            if season._parent_uuid is not None:
                assert season.parent is not None
            assert season.asset_path is not None

    @pytest.mark.asyncio
    async def test_competitive_seasons(self) -> None:
        assert len(self.client.competitive_seasons) > 0
        for cs in self.client.competitive_seasons:
            assert cs is not None
            assert cs.start_time is not None
            assert cs.end_time is not None
            assert cs.season_uuid is not None
            assert cs.competitive_tiers_uuid is not None
            if cs.borders is not None:
                for border in cs.borders:
                    assert border is not None
                    assert border.level is not None
                    assert isinstance(border.level, int)
                    assert border.wins_required is not None
                    assert isinstance(border.wins_required, int)
                    assert border.display_icon is not None
                    if border.small_icon:
                        assert border.small_icon is not None
                    assert border.asset_path is not None
            assert cs.asset_path is not None

    @pytest.mark.asyncio
    async def test_sprays(self) -> None:
        assert len(self.client.sprays) > 0
        for spray in self.client.sprays:
            assert spray is not None
            assert spray.display_name is not None
            if spray.category:
                assert spray.category is not None
            if spray._theme_uuid is not None:
                assert spray.theme is not None
            assert isinstance(spray.is_null_spray(), bool)
            assert spray.display_icon is not None
            if spray.full_icon:
                assert spray.full_icon is not None
            if spray.full_transparent_icon:
                assert spray.full_transparent_icon is not None
            if spray.animation_png:
                assert spray.animation_png is not None
            if spray.animation_gif:
                # TODO: Asset support gif
                assert spray.animation_gif is not None
            assert spray.asset_path is not None
            for level in spray.levels:
                assert level is not None
                assert level.parent is not None
                assert level.spray_level is not None
                assert level.display_name is not None
                assert level.asset_path is not None
                if level.display_icon:
                    assert level.display_icon is not None

    @pytest.mark.asyncio
    async def test_themes(self) -> None:
        assert len(self.client.themes) > 0
        for theme in self.client.themes:
            assert theme is not None
            assert theme.display_name is not None
            if theme.display_icon:
                assert theme.display_icon is not None
            if theme.store_featured_image:
                assert theme.store_featured_image is not None
            assert theme.asset_path is not None

    @pytest.mark.asyncio
    async def test_version(self) -> None:
        version = self.client.version
        assert version.manifest_id is not None
        assert version.branch is not None
        assert version.version is not None
        assert version.build_version is not None
        assert version.engine_version is not None
        assert version.riot_client_version is not None
        assert version.riot_client_build is not None
        assert version.build_date is not None

    @pytest.mark.asyncio
    async def test_weapons(self) -> None:
        assert len(self.client.weapons) > 0
        for weapon in self.client.weapons:
            assert weapon is not None
            assert weapon.display_name is not None
            assert weapon._default_skin_uuid is not None
            assert weapon.display_icon is not None
            assert weapon.kill_stream_icon is not None
            assert weapon.asset_path is not None
            stats = weapon.weapon_stats
            if stats:
                assert stats is not None
                assert stats.fire_rate is not None
                assert stats.magazine_size is not None
                assert stats.run_speed_multiplier is not None
                assert stats.equip_time_seconds is not None
                assert stats.reload_time_seconds is not None
                assert stats.first_bullet_accuracy is not None
                assert stats.wall_penetration is not None
                if stats.feature is not None:
                    assert stats.feature is not None
                if stats.fire_mode is not None:
                    assert stats.fire_mode is not None
                if stats.alt_fire_type is not None:
                    assert stats.alt_fire_type is not None

                # TODO: test all attributes of weapon stats
            if weapon.shop_data:
                assert weapon.shop_data is not None
                # TODO: test all attributes of shop data
            for skin in weapon.skins:
                assert skin is not None
                if skin._theme_uuid is not None:
                    assert skin.theme is not None
                for chroma in skin.chromas:
                    assert chroma is not None
                    assert chroma.parent is not None
                    assert chroma.display_name is not None
                    if chroma.display_icon:
                        assert chroma.display_icon is not None
                    assert chroma.full_render is not None
                    if chroma.swatch:
                        assert chroma.swatch is not None
                    if chroma.streamed_video:
                        assert chroma.streamed_video is not None
                    assert chroma.asset_path is not None

                    # helper method
                    if chroma.display_icon_fix:
                        if skin.theme is not None:
                            assert chroma.theme is not None
                        assert chroma.display_icon_fix is not None

                for index, level in enumerate(skin.levels):
                    assert level is not None
                    assert level.parent is not None
                    assert level.display_name is not None
                    assert level.level_item is not None
                    if level.display_icon:
                        assert level.display_icon is not None
                    if level.streamed_video:
                        assert level.streamed_video is not None
                    assert level.asset_path is not None

                    # helper method
                    assert level.display_icon_fix is not None
                    if skin.theme is not None:
                        assert level.theme is not None
                    assert level.level_number is not None
                    assert isinstance(level.is_level_one(), bool)
                    if index == 0:
                        assert level.is_level_one() is True

    def test_clear(self) -> None:
        self.client.clear()
        assert self.client.version is MISSING
        for key in self.client.cache.__dict__.keys():
            if key.startswith('_') and isinstance(self.client.cache.__dict__[key], dict):
                assert len(self.client.cache.__dict__[key]) == 0

    @pytest.mark.asyncio
    async def test_close(self) -> None:
        await self.client.close()



"""
The MIT License (MIT)

Copyright (c) 2023-present STACiA

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Dict, List, Optional

from .models import (
    Agent,
    Buddy,
    BuddyLevel,
    Bundle,
    Ceremony,
    CompetitiveSeason,
    CompetitiveTier,
    ContentTier,
    Contract,
    Currency,
    Event,
    GameMode,
    GameModeEquippable,
    Gear,
    LevelBorder,
    Map,
    Mission,
    PlayerCard,
    PlayerTitle,
    Season,
    Skin,
    SkinChroma,
    SkinLevel,
    Spray,
    SprayLevel,
    Theme,
    Version,
    Weapon,
)
from .utils import MISSING

if TYPE_CHECKING:
    from .enums import Locale
    from .http import HTTPClient
    from .types import (
        agents,
        buddies,
        bundles,
        ceremonies,
        competitive_tiers,
        content_tiers,
        contracts,
        currencies,
        events,
        gamemodes,
        gear,
        level_borders,
        maps,
        missions,
        player_cards,
        player_titles,
        seasons,
        sprays,
        themes,
        version,
        weapons,
    )

# class BaseCache(ABC):

#     @abstractmethod
#     def get(self):
#         raise NotImplementedError

#     @abstractmethod
#     def find(self):
#         raise NotImplementedError

_log = logging.getLogger(__name__)


class CacheState:
    def __init__(self, *, locale: Locale, http: HTTPClient, to_file: bool = False) -> None:
        self.locale = locale
        self.http = http
        self.cache: bool = True
        self._to_file: bool = to_file
        # -
        self._agents: Dict[str, Agent] = {}
        self._buddies: Dict[str, Buddy] = {}
        self._buddy_levels: Dict[str, BuddyLevel] = {}
        self._bundles: Dict[str, Bundle] = {}
        self._ceremonies: Dict[str, Ceremony] = {}
        self._competitive_tiers: Dict[str, CompetitiveTier] = {}
        self._content_tiers: Dict[str, ContentTier] = {}
        self._contracts: Dict[str, Contract] = {}
        self._currencies: Dict[str, Currency] = {}
        self._events: Dict[str, Event] = {}
        self._game_modes: Dict[str, GameMode] = {}
        self._game_mode_equippables: Dict[str, GameModeEquippable] = {}
        self._gear: Dict[str, Gear] = {}
        self._level_borders: Dict[str, LevelBorder] = {}
        self._maps: Dict[str, Map] = {}
        self._missions: Dict[str, Mission] = {}
        self._player_cards: Dict[str, PlayerCard] = {}
        self._player_titles: Dict[str, PlayerTitle] = {}
        self._seasons: Dict[str, Season] = {}
        self._competitive_seasons: Dict[str, CompetitiveSeason] = {}
        self._sprays: Dict[str, Spray] = {}
        self._spray_levels: Dict[str, SprayLevel] = {}
        self._themes: Dict[str, Theme] = {}
        self._version: Version = MISSING
        self._weapons: Dict[str, Weapon] = {}
        self._skins: Dict[str, Skin] = {}
        self._skin_chromas: Dict[str, SkinChroma] = {}
        self._skin_levels: Dict[str, SkinLevel] = {}

    async def init(self) -> None:
        tasks = [
            self.http.get_agents,
            self.http.get_buddies,
            self.http.get_bundles,
            self.http.get_ceremonies,
            self.http.get_competitive_tiers,
            self.http.get_content_tiers,
            self.http.get_contracts,
            self.http.get_currencies,
            self.http.get_events,
            self.http.get_game_modes,
            self.http.get_game_mode_equippables,
            self.http.get_gear,
            self.http.get_level_borders,
            self.http.get_maps,
            self.http.get_missions,
            self.http.get_player_cards,
            self.http.get_player_titles,
            self.http.get_seasons,
            self.http.get_competitive_seasons,
            self.http.get_sprays,
            self.http.get_themes,
            self.http.get_weapons,
            self.http.get_version,
        ]
        # if self._to_file:

        #     # read from file
        #     for func in tasks:
        #         funcname = func.__name__.split('_')[1:]
        #         filename = '_'.join(funcname) + '.json'
        #         # add_to_cache = getattr(self, f'add_{funcname}')
        #         # if os.path.exists(filename):
        #         #     with open(filename, 'r') as f:
        #         #         data = json.load(f)
        #         #         continue
        #         # else:
        #         #     data = await func()
        #         #     self._add_to_file(data, filename)
        #         # add_to_cache = getattr(self, f'add_{funcname}')
        # else:
        # results = await self._fetch_data_from_api(tasks)

        results = await asyncio.gather(*(task() for task in tasks))
        for func, result in zip(tasks, results):
            assert result is not None
            funcname = func.__name__.split('_')[1:]
            funcname = '_'.join(funcname)
            parse_func = getattr(self, f'_add_{funcname}')
            parse_func(result)

    # async def _fetch_data_from_api(self, tasks: List[Callable[..., Coroutine[Any, Any, Any]]]) -> List[Any]:
    #     results = await asyncio.gather(*[task() for task in tasks])
    #     return results
    # for func, result in zip(tasks, results):
    #     assert result is not None
    #     funcname = func.__name__.split('_')[1:]
    #     funcname = '_'.join(funcname)
    #     parse_func = getattr(self, f'add_{funcname}')
    #     parse_func(result)

    # def _add_to_file(self, data: Any, filename: str) -> None:
    #     with open(filename, 'w') as f:
    #         json.dump(data, f, indent=4)

    def clear(self) -> None:
        for key in self.__dict__.keys():
            if key.startswith('_') and isinstance(self.__dict__[key], dict):
                self.__dict__[key].clear()
        self._version = MISSING

    # agents

    @property
    def agents(self) -> List[Agent]:
        return list(self._agents.values())

    def get_agent(self, uuid: Optional[str], /) -> Optional[Agent]:
        return self._agents.get(uuid)  # type: ignore

    def store_agent(self, data: agents.Agent) -> Agent:
        agent_id = data['uuid']
        self._agents[agent_id] = agent = Agent(state=self, data=data)
        return agent

    def _add_agents(self, data: agents.Agents) -> None:
        agent_data = data['data']
        for agent in agent_data:
            existing = self.get_agent(agent['uuid'])
            if existing is None:
                self.store_agent(agent)

    async def fetch_agents(self) -> List[Agent]:
        data = await self.http.get_agents()
        self._add_agents(data)
        return self.agents

    # buddies

    @property
    def buddies(self) -> List[Buddy]:
        return list(self._buddies.values())

    @property
    def buddy_levels(self) -> List[BuddyLevel]:
        return list(self._buddy_levels.values())

    def get_buddy(self, uuid: Optional[str], /) -> Optional[Buddy]:
        return self._buddies.get(uuid)  # type: ignore

    def get_buddy_level(self, uuid: Optional[str], /) -> Optional[BuddyLevel]:
        return self._buddy_levels.get(uuid)  # type: ignore

    def store_buddy(self, data: buddies.Buddy) -> Buddy:
        buddy_id = data['uuid']
        self._buddies[buddy_id] = buddy = Buddy(state=self, data=data)
        for level in buddy.levels:
            self._buddy_levels[level.uuid] = level
        return buddy

    def _add_buddies(self, data: buddies.Buddies) -> None:
        buddy_data = data['data']
        for buddy in buddy_data:
            buddy_existing = self.get_buddy(buddy['uuid'])
            if buddy_existing is None:
                self.store_buddy(buddy)

    async def fetch_buddies(self) -> List[Buddy]:
        data = await self.http.get_buddies()
        self._add_buddies(data)
        return self.buddies

    # bundles

    @property
    def bundles(self) -> List[Bundle]:
        return list(self._bundles.values())

    def get_bundle(self, uuid: Optional[str], /) -> Optional[Bundle]:
        return self._bundles.get(uuid)  # type: ignore

    def store_bundle(self, data: bundles.Bundle) -> Bundle:
        bundle_id = data['uuid']
        self._bundles[bundle_id] = bundle = Bundle(state=self, data=data)
        return bundle

    def _add_bundles(self, data: bundles.Bundles) -> None:
        bundle_data = data['data']
        for bundle in bundle_data:
            bundle_existing = self.get_bundle(bundle['uuid'])
            if bundle_existing is None:
                self.store_bundle(bundle)

    async def fetch_bundles(self) -> List[Bundle]:
        data = await self.http.get_bundles()
        self._add_bundles(data)
        return self.bundles

    # ceremonies

    @property
    def ceremonies(self) -> List[Ceremony]:
        return list(self._ceremonies.values())

    def get_ceremony(self, uuid: Optional[str], /) -> Optional[Ceremony]:
        return self._ceremonies.get(uuid)  # type: ignore

    def store_ceremony(self, data: ceremonies.Ceremony) -> Ceremony:
        ceremony_id = data['uuid']
        self._ceremonies[ceremony_id] = ceremony = Ceremony(state=self, data=data)
        return ceremony

    def _add_ceremonies(self, data: ceremonies.Ceremonies) -> None:
        ceremony_data = data['data']
        for ceremony in ceremony_data:
            ceremony_existing = self.get_ceremony(ceremony['uuid'])
            if ceremony_existing is None:
                self.store_ceremony(ceremony)

    async def fetch_ceremonies(self) -> List[Ceremony]:
        data = await self.http.get_ceremonies()
        self._add_ceremonies(data)
        return self.ceremonies

    # competitive tiers

    @property
    def competitive_tiers(self) -> List[CompetitiveTier]:
        return list(self._competitive_tiers.values())

    def get_competitive_tier(self, uuid: Optional[str], /) -> Optional[CompetitiveTier]:
        return self._competitive_tiers.get(uuid)  # type: ignore

    def store_competitive_tier(self, data: competitive_tiers.CompetitiveTier) -> CompetitiveTier:
        competitive_tier_id = data['uuid']
        self._competitive_tiers[competitive_tier_id] = competitive_tier = CompetitiveTier(state=self, data=data)
        return competitive_tier

    def _add_competitive_tiers(self, data: competitive_tiers.CompetitiveTiers) -> None:
        competitive_tier_data = data['data']
        for competitive_tier in competitive_tier_data:
            competitive_tier_existing = self.get_competitive_tier(competitive_tier['uuid'])
            if competitive_tier_existing is None:
                self.store_competitive_tier(competitive_tier)

    async def fetch_competitive_tiers(self) -> List[CompetitiveTier]:
        data = await self.http.get_competitive_tiers()
        self._add_competitive_tiers(data)
        return self.competitive_tiers

    # content tiers

    @property
    def content_tiers(self) -> List[ContentTier]:
        return list(self._content_tiers.values())

    def get_content_tier(self, uuid: Optional[str], /) -> Optional[ContentTier]:
        return self._content_tiers.get(uuid)  # type: ignore

    def store_content_tier(self, data: content_tiers.ContentTier) -> ContentTier:
        content_tier_id = data['uuid']
        self._content_tiers[content_tier_id] = content_tier = ContentTier(state=self, data=data)
        return content_tier

    def _add_content_tiers(self, data: content_tiers.ContentTiers) -> None:
        content_tier_data = data['data']
        for content_tier in content_tier_data:
            content_tier_existing = self.get_content_tier(content_tier['uuid'])
            if content_tier_existing is None:
                self.store_content_tier(content_tier)

    async def fetch_content_tiers(self) -> List[ContentTier]:
        data = await self.http.get_content_tiers()
        self._add_content_tiers(data)
        return self.content_tiers

    # contracts

    @property
    def contracts(self) -> List[Contract]:
        return list(self._contracts.values())

    def get_contract(self, uuid: Optional[str], /) -> Optional[Contract]:
        return self._contracts.get(uuid)  # type: ignore

    def store_contract(self, data: contracts.Contract) -> Contract:
        contract_id = data['uuid']
        self._contracts[contract_id] = contract = Contract(state=self, data=data)
        return contract

    def _add_contracts(self, data: contracts.Contracts) -> None:
        contract_data = data['data']
        for contract in contract_data:
            contract_existing = self.get_contract(contract['uuid'])
            if contract_existing is None:
                self.store_contract(contract)

    async def fetch_contracts(self) -> List[Contract]:
        data = await self.http.get_contracts()
        self._add_contracts(data)
        return self.contracts

    # currencies

    @property
    def currencies(self) -> List[Currency]:
        return list(self._currencies.values())

    def get_currency(self, uuid: Optional[str], /) -> Optional[Currency]:
        return self._currencies.get(uuid)  # type: ignore

    def store_currency(self, data: currencies.Currency) -> Currency:
        currency_id = data['uuid']
        self._currencies[currency_id] = currency = Currency(state=self, data=data)
        return currency

    def _add_currencies(self, data: currencies.Currencies) -> None:
        currency_data = data['data']
        for currency in currency_data:
            currency_existing = self.get_currency(currency['uuid'])
            if currency_existing is None:
                self.store_currency(currency)

    async def fetch_currencies(self) -> List[Currency]:
        data = await self.http.get_currencies()
        self._add_currencies(data)
        return self.currencies

    # events

    @property
    def events(self) -> List[Event]:
        return list(self._events.values())

    def get_event(self, uuid: Optional[str], /) -> Optional[Event]:
        return self._events.get(uuid)  # type: ignore

    def store_event(self, data: events.Event) -> Event:
        event_id = data['uuid']
        self._events[event_id] = event = Event(state=self, data=data)
        return event

    def _add_events(self, data: events.Events) -> None:
        event_data = data['data']
        for event in event_data:
            event_existing = self.get_event(event['uuid'])
            if event_existing is None:
                self.store_event(event)

    async def fetch_events(self) -> List[Event]:
        data = await self.http.get_events()
        self._add_events(data)
        return self.events

    # game modes

    @property
    def game_modes(self) -> List[GameMode]:
        return list(self._game_modes.values())

    @property
    def game_mode_equippables(self) -> List[GameModeEquippable]:
        return list(self._game_mode_equippables.values())

    def get_game_mode(self, uuid: Optional[str], /) -> Optional[GameMode]:
        return self._game_modes.get(uuid)  # type: ignore

    def get_game_mode_equippable(self, uuid: Optional[str], /) -> Optional[GameModeEquippable]:
        return self._game_mode_equippables.get(uuid)  # type: ignore

    def store_game_mode(self, data: gamemodes.GameMode) -> GameMode:
        game_mode_id = data['uuid']
        self._game_modes[game_mode_id] = game_mode = GameMode(state=self, data=data)
        return game_mode

    def store_game_mode_equippable(self, data: gamemodes.GameModeEquippable) -> GameModeEquippable:
        game_mode_equippable_id = data['uuid']
        self._game_mode_equippables[game_mode_equippable_id] = game_mode_equippable = GameModeEquippable(
            state=self, data=data
        )
        return game_mode_equippable

    def _add_game_modes(self, data: gamemodes.GameModes) -> None:
        game_mode_data = data['data']
        for game_mode in game_mode_data:
            game_mode_existing = self.get_game_mode(game_mode['uuid'])
            if game_mode_existing is None:
                self.store_game_mode(game_mode)

    async def fetch_game_modes(self) -> List[GameMode]:
        data = await self.http.get_game_modes()
        self._add_game_modes(data)
        return self.game_modes

    def _add_game_mode_equippables(self, data: gamemodes.GameModeEquippables) -> None:
        game_mode_equippable_data = data['data']
        for game_mode_equippable in game_mode_equippable_data:
            game_mode_equippable_existing = self.get_game_mode_equippable(game_mode_equippable['uuid'])
            if game_mode_equippable_existing is None:
                self.store_game_mode_equippable(game_mode_equippable)

    async def fetch_game_mode_equippables(self) -> List[GameModeEquippable]:
        data = await self.http.get_game_mode_equippables()
        self._add_game_mode_equippables(data)
        return self.game_mode_equippables

    # gear

    @property
    def gear(self) -> List[Gear]:
        return list(self._gear.values())

    def get_gear(self, uuid: Optional[str], /) -> Optional[Gear]:
        return self._gear.get(uuid)  # type: ignore

    def store_gear(self, data: gear.Gear_) -> Gear:
        gear_id = data['uuid']
        self._gear[gear_id] = gear = Gear(state=self, data=data)
        return gear

    def _add_gear(self, data: gear.Gear) -> None:
        gear_data = data['data']
        for gear_ in gear_data:
            gear_existing = self.get_gear(gear_['uuid'])
            if gear_existing is None:
                self.store_gear(gear_)

    async def fetch_gear(self) -> List[Gear]:
        data = await self.http.get_gear()
        self._add_gear(data)
        return self.gear

    # level borders

    @property
    def level_borders(self) -> List[LevelBorder]:
        return list(self._level_borders.values())

    def get_level_border(self, uuid: Optional[str], /) -> Optional[LevelBorder]:
        return self._level_borders.get(uuid)  # type: ignore

    def store_level_border(self, data: level_borders.LevelBorder) -> LevelBorder:
        level_border_id = data['uuid']
        self._level_borders[level_border_id] = level_border = LevelBorder(state=self, data=data)
        return level_border

    def _add_level_borders(self, data: level_borders.LevelBorders) -> None:
        level_border_data = data['data']
        for level_border in level_border_data:
            level_border_existing = self.get_level_border(level_border['uuid'])
            if level_border_existing is None:
                self.store_level_border(level_border)

    async def fetch_level_borders(self) -> List[LevelBorder]:
        data = await self.http.get_level_borders()
        self._add_level_borders(data)
        return self.level_borders

    # maps

    @property
    def maps(self) -> List[Map]:
        return list(self._maps.values())

    def get_map(self, uuid: Optional[str], /) -> Optional[Map]:
        return self._maps.get(uuid)  # type: ignore

    def store_map(self, data: maps.Map) -> Map:
        map_id = data['uuid']
        self._maps[map_id] = map_ = Map(state=self, data=data)
        return map_

    def _add_maps(self, data: maps.Maps) -> None:
        map_data = data['data']
        for map_ in map_data:
            map_existing = self.get_map(map_['uuid'])
            if map_existing is None:
                self.store_map(map_)

    async def fetch_maps(self) -> List[Map]:
        data = await self.http.get_maps()
        self._add_maps(data)
        return self.maps

    # missions

    @property
    def missions(self) -> List[Mission]:
        return list(self._missions.values())

    def get_mission(self, uuid: Optional[str], /) -> Optional[Mission]:
        return self._missions.get(uuid)  # type: ignore

    def store_mission(self, data: missions.Mission) -> Mission:
        mission_id = data['uuid']
        self._missions[mission_id] = mission = Mission(state=self, data=data)
        return mission

    def _add_missions(self, data: missions.Missions) -> None:
        mission_data = data['data']
        for mission in mission_data:
            mission_existing = self.get_mission(mission['uuid'])
            if mission_existing is None:
                self.store_mission(mission)

    async def fetch_missions(self) -> List[Mission]:
        data = await self.http.get_missions()
        self._add_missions(data)
        return self.missions

    # player_cards

    @property
    def player_cards(self) -> List[PlayerCard]:
        return list(self._player_cards.values())

    def get_player_card(self, uuid: Optional[str], /) -> Optional[PlayerCard]:
        return self._player_cards.get(uuid)  # type: ignore

    def store_player_card(self, data: player_cards.PlayerCard) -> PlayerCard:
        player_card_id = data['uuid']
        self._player_cards[player_card_id] = player_card = PlayerCard(state=self, data=data)
        return player_card

    def _add_player_cards(self, data: player_cards.PlayerCards) -> None:
        player_card_data = data['data']
        for player_card in player_card_data:
            player_card_existing = self.get_player_card(player_card['uuid'])
            if player_card_existing is None:
                self.store_player_card(player_card)

    async def fetch_player_cards(self) -> List[PlayerCard]:
        data = await self.http.get_player_cards()
        self._add_player_cards(data)
        return self.player_cards

    # player_titles

    @property
    def player_titles(self) -> List[PlayerTitle]:
        return list(self._player_titles.values())

    def get_player_title(self, uuid: Optional[str], /) -> Optional[PlayerTitle]:
        return self._player_titles.get(uuid)  # type: ignore

    def store_player_title(self, data: player_titles.PlayerTitle) -> PlayerTitle:
        player_title_id = data['uuid']
        self._player_titles[player_title_id] = player_title = PlayerTitle(state=self, data=data)
        return player_title

    def _add_player_titles(self, data: player_titles.PlayerTitles) -> None:
        player_title_data = data['data']
        for player_title in player_title_data:
            player_title_existing = self.get_player_title(player_title['uuid'])
            if player_title_existing is None:
                self.store_player_title(player_title)

    async def fetch_player_titles(self) -> List[PlayerTitle]:
        data = await self.http.get_player_titles()
        self._add_player_titles(data)
        return self.player_titles

    # seasons

    @property
    def seasons(self) -> List[Season]:
        return list(self._seasons.values())

    def get_season(self, uuid: Optional[str], /) -> Optional[Season]:
        return self._seasons.get(uuid)  # type: ignore

    def store_season(self, data: seasons.Season) -> Season:
        season_id = data['uuid']
        self._seasons[season_id] = season = Season(state=self, data=data)
        return season

    def _add_seasons(self, data: seasons.Seasons) -> None:
        season_data = data['data']
        for season in season_data:
            season_existing = self.get_season(season['uuid'])
            if season_existing is None:
                self.store_season(season)

    async def fetch_seasons(self) -> List[Season]:
        data = await self.http.get_seasons()
        self._add_seasons(data)
        return self.seasons

    @property
    def competitive_seasons(self) -> List[CompetitiveSeason]:
        return list(self._competitive_seasons.values())

    def get_competitive_season(self, uuid: Optional[str], /) -> Optional[CompetitiveSeason]:
        return self._competitive_seasons.get(uuid)  # type: ignore

    def store_competitive_season(self, data: seasons.CompetitiveSeason) -> CompetitiveSeason:
        season_id = data['uuid']
        self._competitive_seasons[season_id] = season = CompetitiveSeason(state=self, data=data)
        return season

    def _add_competitive_seasons(self, data: seasons.CompetitiveSeasons) -> None:
        season_data = data['data']
        for season in season_data:
            season_existing = self.get_competitive_season(season['uuid'])
            if season_existing is None:
                self.store_competitive_season(season)

    async def fetch_competitive_seasons(self) -> List[CompetitiveSeason]:
        data = await self.http.get_competitive_seasons()
        self._add_competitive_seasons(data)
        return self.competitive_seasons

    # sprays

    @property
    def sprays(self) -> List[Spray]:
        return list(self._sprays.values())

    def get_spray(self, uuid: Optional[str], /) -> Optional[Spray]:
        return self._sprays.get(uuid)  # type: ignore

    @property
    def spray_levels(self) -> List[SprayLevel]:
        return list(self._spray_levels.values())

    def get_spray_level(self, uuid: Optional[str], /) -> Optional[SprayLevel]:
        return self._spray_levels.get(uuid)  # type: ignore

    def store_spray(self, data: sprays.Spray) -> Spray:
        spray_id = data['uuid']
        self._sprays[spray_id] = spray = Spray(state=self, data=data)
        for level in spray.levels:
            self._spray_levels[level.uuid] = level
        return spray

    def _add_sprays(self, data: sprays.Sprays) -> None:
        spray_data = data['data']
        for spray in spray_data:
            spray_existing = self.get_spray(spray['uuid'])
            if spray_existing is None:
                self.store_spray(spray)

    async def fetch_sprays(self) -> List[Spray]:
        data = await self.http.get_sprays()
        self._add_sprays(data)
        return self.sprays

    # themes

    @property
    def themes(self) -> List[Theme]:
        return list(self._themes.values())

    def get_theme(self, uuid: Optional[str], /) -> Optional[Theme]:
        return self._themes.get(uuid)  # type: ignore

    def store_theme(self, data: themes.Theme) -> Theme:
        theme_id = data['uuid']
        self._themes[theme_id] = theme = Theme(state=self, data=data)
        return theme

    def _add_themes(self, data: themes.Themes) -> None:
        theme_data = data['data']
        for theme in theme_data:
            theme_existing = self.get_theme(theme['uuid'])
            if theme_existing is None:
                self.store_theme(theme)

    async def fetch_themes(self) -> List[Theme]:
        data = await self.http.get_themes()
        self._add_themes(data)
        return self.themes

    # version

    @property
    def version(self) -> Version:
        return self._version

    def _add_version(self, data: version.Version) -> None:
        self._version = Version(state=self, data=data['data'])

    async def fetch_version(self) -> Version:
        data = await self.http.get_version()
        self._add_version(data)
        return self.version

    # weapons

    @property
    def weapons(self) -> List[Weapon]:
        return list(self._weapons.values())

    def get_weapon(self, uuid: Optional[str], /) -> Optional[Weapon]:
        return self._weapons.get(uuid)  # type: ignore

    @property
    def skins(self) -> List[Skin]:
        return list(self._skins.values())

    def get_skin(self, uuid: Optional[str], /) -> Optional[Skin]:
        return self._skins.get(uuid)  # type: ignore

    @property
    def skin_chromas(self) -> List[SkinChroma]:
        return list(self._skin_chromas.values())

    def get_skin_chroma(self, uuid: Optional[str], /) -> Optional[SkinChroma]:
        return self._skin_chromas.get(uuid)  # type: ignore

    @property
    def skin_levels(self) -> List[SkinLevel]:
        return list(self._skin_levels.values())

    def get_skin_level(self, uuid: Optional[str], /) -> Optional[SkinLevel]:
        return self._skin_levels.get(uuid)  # type: ignore

    def store_weapon(self, data: weapons.Weapon) -> Weapon:
        weapon_id = data['uuid']
        self._weapons[weapon_id] = weapon = Weapon(state=self, data=data)
        for skin in weapon.skins:
            self._skins[skin.uuid] = skin
            for chroma in skin.chromas:
                self._skin_chromas[chroma.uuid] = chroma
            for level in skin.levels:
                self._skin_levels[level.uuid] = level
        return weapon

    def _add_weapons(self, data: weapons.Weapons) -> None:
        weapon_data = data['data']
        for weapon in weapon_data:
            weapon_existing = self.get_weapon(weapon['uuid'])
            if weapon_existing is None:
                self.store_weapon(weapon)

    async def fetch_weapons(self) -> List[Weapon]:
        data = await self.http.get_weapons()
        self._add_weapons(data)
        return self.weapons



"""
The MIT License (MIT)

Copyright (c) 2023-present STACiA

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Optional, Union

from .. import utils
from ..asset import Asset
from ..enums import Locale
from ..localization import Localization
from .abc import BaseModel

if TYPE_CHECKING:
    from ..cache import CacheState
    from ..types.competitive_tiers import CompetitiveTier as CompetitiveTierPayload, Tier as TierPayload

# fmt: off
__all__ = (
    'CompetitiveTier',
    'Tier',
)
# fmt: on


class Tier:
    def __init__(self, state: CacheState, data: TierPayload) -> None:
        self._state: CacheState = state
        self.tier: int = data['tier']
        self._name: Union[str, Dict[str, str]] = data['tierName']
        self._division: Optional[str] = data['division']
        self._division_name: Union[str, Dict[str, str]] = data['divisionName']
        self.color: str = data['color']
        self.background_color: str = data['backgroundColor']
        self._small_icon: Optional[str] = data['smallIcon']
        self._large_icon: Optional[str] = data['largeIcon']
        self._rank_triangle_down_icon: Optional[str] = data['rankTriangleDownIcon']
        self._rank_triangle_up_icon: Optional[str] = data['rankTriangleUpIcon']
        self._name_locale: Localization = Localization(self._name, locale=self._state.locale)
        self._division_name_localized: Localization = Localization(self._division_name, locale=self._state.locale)

    def __str__(self) -> str:
        return self.name.locale

    def __repr__(self) -> str:
        return f'<Tier tier={self.tier!r} name={self.name!r} division={self.division!r}>'

    def __hash__(self) -> int:
        return hash(self.tier)

    def __eq__(self, other) -> bool:
        return isinstance(other, Tier) and self.tier == other.tier

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)

    def __lt__(self, other: object) -> bool:
        return isinstance(other, Tier) and self.tier < other.tier

    def __le__(self, other: object) -> bool:
        return isinstance(other, Tier) and self.tier <= other.tier

    def __gt__(self, other: object) -> bool:
        return isinstance(other, Tier) and self.tier > other.tier

    def __ge__(self, other: object) -> bool:
        return isinstance(other, Tier) and self.tier >= other.tier

    def display_name_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        return self._name_locale.from_locale(locale)

    def division_name_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        return self._division_name_localized.from_locale(locale)

    @property
    def name(self) -> Localization:
        """:class: `str` Returns the tier's name."""
        return self._name_locale

    @property
    def display_name(self) -> Localization:
        """:class: `Localization` alias for :attr:`name`"""
        return self.name

    @property
    def division(self) -> Optional[str]:
        """:class: `str` Returns the tier's division."""
        return utils.removeprefix(self._division, 'ECompetitiveDivision::') if self._division else None

    @property
    def division_name(self) -> Localization:
        """:class: `str` Returns the tier's division."""
        return self._division_name_localized

    @property
    def small_icon(self) -> Optional[Asset]:
        """:class: `Asset` Returns the tier's small icon."""
        if self._small_icon is None:
            return None
        return Asset(self._state, self._small_icon)

    @property
    def large_icon(self) -> Optional[Asset]:
        """:class: `Asset` Returns the tier's large icon."""
        if self._large_icon is None:
            return None
        return Asset(self._state, self._large_icon)

    @property
    def rank_triangle_down_icon(self) -> Optional[Asset]:
        """:class: `Asset` Returns the tier's rank triangle down icon."""
        if self._rank_triangle_down_icon is None:
            return None
        return Asset(self._state, self._rank_triangle_down_icon)

    @property
    def rank_triangle_up_icon(self) -> Optional[Asset]:
        """:class: `Asset` Returns the tier's rank triangle up icon."""
        if self._rank_triangle_up_icon is None:
            return None
        return Asset(self._state, self._rank_triangle_up_icon)


class CompetitiveTier(BaseModel):
    def __init__(self, state: CacheState, data: CompetitiveTierPayload) -> None:
        super().__init__(data['uuid'])
        self._state: CacheState = state
        self.asset_object_name: str = data['assetObjectName']
        self._tiers: Dict[int, Tier] = {tier['tier']: Tier(state=self._state, data=tier) for tier in data['tiers']}
        self.asset_path: str = data['assetPath']

    def __str__(self) -> str:
        return self.asset_object_name

    def __repr__(self) -> str:
        return f'<CompetitiveTier asset_object_name={self.asset_object_name!r}>'

    @property
    def tiers(self) -> List[Tier]:
        """:class: `list` Returns the competitive tier's tiers."""
        return list(self._tiers.values())

    def get_tier(self, tier: int) -> Optional[Tier]:
        """Returns the tier with the given tier number."""
        return self._tiers.get(tier)



"""
The MIT License (MIT)

Copyright (c) 2023-present STACiA

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from ..asset import Asset
from ..enums import AbilityType, Locale, try_enum
from ..localization import Localization
from .abc import BaseModel

if TYPE_CHECKING:
    from ..cache import CacheState
    from ..types.agents import (
        Ability as AbilityPayload,
        Agent as AgentPayload,
        Media as MediaPayload,
        Role as RolePayload,
        VoiceLine as VoiceLinePayload,
    )

# fmt: off
__all__ = (
    'Ability',
    'Agent',
    'Media',
    'Role',
    'VoiceLine',
    'VoiceLineLocalization',
)
# fmt: on


class Role(BaseModel):
    def __init__(self, state: CacheState, data: RolePayload) -> None:
        super().__init__(data['uuid'])
        self._state = state
        self._display_name: Union[str, Dict[str, str]] = data['displayName']
        self._description: Union[str, Dict[str, str]] = data['description']
        self._display_icon: str = data['displayIcon']
        self.asset_path: str = data['assetPath']
        self._display_name_localized: Localization = Localization(self._description, locale=self._state.locale)
        self._description_localized: Localization = Localization(self._display_name, locale=self._state.locale)

    def __repr__(self) -> str:
        return f'<Role display_name={self.display_name!r}>'

    def __str__(self) -> str:
        return self.display_name.locale

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Role) and other.uuid == self.uuid

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(self.uuid)

    def display_name_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        return self._display_name_localized.from_locale(locale)

    def description_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        return self._description_localized.from_locale(locale)

    @property
    def display_name(self) -> Localization:
        """:class: `str` Returns the agent role's name."""
        return self._display_name_localized

    @property
    def description(self) -> Localization:
        """:class: `str` Returns the agent role's description."""
        return self._description_localized

    @property
    def display_icon(self) -> Asset:
        """:class: `Asset` Returns the agent role's display icon."""
        return Asset._from_url(state=self._state, url=self._display_icon)


class Ability:
    def __init__(self, state: CacheState, data: AbilityPayload, agent: Agent) -> None:
        self._state = state
        self.agent: Agent = agent
        self.slot: AbilityType = try_enum(AbilityType, data['slot'])
        self._display_name: Union[str, Dict[str, str]] = data['displayName']
        self._description: Union[str, Dict[str, str]] = data['description']
        self._display_icon: Optional[str] = data['displayIcon']
        self._display_name_localized: Localization = Localization(self._display_name, locale=self._state.locale)
        self._description_localized: Localization = Localization(self._description, locale=self._state.locale)

    def __repr__(self) -> str:
        return f'<Ability display_name={self.display_name!r}>'

    def __str__(self) -> str:
        return self.display_name.locale

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Ability) and other.slot == self.slot

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def display_name_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        return self._display_name_localized.from_locale(locale)

    def description_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        return self._description_localized.from_locale(locale)

    @property
    def display_name(self) -> Localization:
        """:class: `str` Returns the agent role's name."""
        return self._display_name_localized

    @property
    def description(self) -> Localization:
        """:class: `str` Returns the agent role's description."""
        return self._description_localized

    @property
    def display_icon(self) -> Optional[Asset]:
        """:class: `Asset` Returns the agent role's display icon."""
        if self._display_icon is None:
            return None
        return Asset._from_url(state=self._state, url=self._display_icon)


class Media:
    def __init__(self, data: MediaPayload) -> None:
        self.id: int = data['id']
        self.wwise: str = data['wwise']
        self.wave: str = data['wave']

    def __repr__(self) -> str:
        return f'<Media id={self.id!r} wwise={self.wwise!r} wave={self.wave!r}>'

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Media) and other.id == self.id

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(self.id)


class VoiceLine:
    def __init__(self, data: VoiceLinePayload) -> None:
        self.min_duration: float = data['minDuration']
        self.max_duration: float = data['maxDuration']
        self.media_list: List[Media] = [Media(media) for media in data['mediaList']]

    def __repr__(self) -> str:
        return f'<AgentVoiceLine min_duration={self.min_duration!r} max_duration={self.max_duration!r}>'

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, VoiceLine)
            and other.min_duration == self.min_duration
            and other.max_duration == self.max_duration
        )

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash((self.min_duration, self.max_duration))


class VoiceLineLocalization:
    def __init__(
        self,
        untranslated: Any,
        locale: Locale = Locale.american_english,
    ) -> None:
        self.ut = untranslated
        self._locale = locale

        # locale code
        self.ar_AE: Optional[VoiceLine] = VoiceLine(self.ut['ar-AE']) if self.ut.get('ar-AE') else None
        self.de_DE: Optional[VoiceLine] = VoiceLine(self.ut['de-DE']) if self.ut.get('de-DE') else None
        self.en_US: Optional[VoiceLine] = VoiceLine(self.ut['en-US']) if self.ut.get('en-US') else None
        self.es_ES: Optional[VoiceLine] = VoiceLine(self.ut['es-ES']) if self.ut.get('es-ES') else None
        self.es_MX: Optional[VoiceLine] = VoiceLine(self.ut['es-MX']) if self.ut.get('es-MX') else None
        self.fr_FR: Optional[VoiceLine] = VoiceLine(self.ut['fr-FR']) if self.ut.get('fr-FR') else None
        self.id_ID: Optional[VoiceLine] = VoiceLine(self.ut['id-ID']) if self.ut.get('id-ID') else None
        self.it_IT: Optional[VoiceLine] = VoiceLine(self.ut['it-IT']) if self.ut.get('it-IT') else None
        self.ja_JP: Optional[VoiceLine] = VoiceLine(self.ut['ja-JP']) if self.ut.get('ja-JP') else None
        self.ko_KR: Optional[VoiceLine] = VoiceLine(self.ut['ko-KR']) if self.ut.get('ko-KR') else None
        self.pl_PL: Optional[VoiceLine] = VoiceLine(self.ut['pl-PL']) if self.ut.get('pl-PL') else None
        self.pt_BR: Optional[VoiceLine] = VoiceLine(self.ut['pt-BR']) if self.ut.get('pt-BR') else None
        self.ru_RU: Optional[VoiceLine] = VoiceLine(self.ut['ru-RU']) if self.ut.get('ru-RU') else None
        self.th_TH: Optional[VoiceLine] = VoiceLine(self.ut['th-TH']) if self.ut.get('th-TH') else None
        self.tr_TR: Optional[VoiceLine] = VoiceLine(self.ut['tr-TR']) if self.ut.get('tr-TR') else None
        self.vi_VN: Optional[VoiceLine] = VoiceLine(self.ut['vi-VN']) if self.ut.get('vi-VN') else None
        self.zh_CN: Optional[VoiceLine] = VoiceLine(self.ut['zh-CN']) if self.ut.get('zh-CN') else None
        self.zh_TW: Optional[VoiceLine] = VoiceLine(self.ut['zh-TW']) if self.ut.get('zh-TW') else None

    def __repr__(self) -> str:
        attrs = [
            ('ar_AE', self.ar_AE),
            ('de_DE', self.de_DE),
            ('en_US', self.en_US),
            ('es_ES', self.es_ES),
            ('es_MX', self.es_MX),
            ('fr_FR', self.fr_FR),
            ('id_ID', self.id_ID),
            ('it_IT', self.it_IT),
            ('ja_JP', self.ja_JP),
            ('ko_KR', self.ko_KR),
            ('pl_PL', self.pl_PL),
            ('pt_BR', self.pt_BR),
            ('ru_RU', self.ru_RU),
            ('th_TH', self.th_TH),
            ('tr_TR', self.tr_TR),
            ('vi_VN', self.vi_VN),
            ('zh_CN', self.zh_CN),
            ('zh_TW', self.zh_TW),
        ]
        joined = ' '.join('%s=%r' % t for t in attrs)
        return f'<{self.__class__.__name__} {joined}>'

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Localization) and self.ut == other.untranslated

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(self.ut)

    def _dict(self) -> Dict[Locale, Optional[VoiceLine]]:
        """:class:`Dict[Locale, VoiceLine]`: Returns all locales as a dictionary."""
        voices = {
            Locale.arabic: self.ar_AE,
            Locale.german: self.de_DE,
            Locale.american_english: self.en_US,
            Locale.british_english: self.en_US,
            Locale.english: self.en_US,
            Locale.spain_spanish: self.es_ES,
            Locale.spanish_mexican: self.es_MX,
            Locale.french: self.fr_FR,
            Locale.indonesian: self.id_ID,
            Locale.italian: self.it_IT,
            Locale.japanese: self.ja_JP,
            Locale.korean: self.ko_KR,
            Locale.polish: self.pl_PL,
            Locale.brazil_portuguese: self.pt_BR,
            Locale.russian: self.ru_RU,
            Locale.thai: self.th_TH,
            Locale.turkish: self.tr_TR,
            Locale.vietnamese: self.vi_VN,
            Locale.chinese: self.zh_CN,
            Locale.taiwan_chinese: self.zh_TW,
        }
        return voices

    def all(self) -> List[VoiceLine]:
        """:class:`List[VoiceLine]`: Returns all locales as a list."""
        return list([v for v in self._dict().values() if v is not None])

    @property
    def voice_locale(self) -> Optional[VoiceLine]:
        """:class:`Optional[VoiceLine]`: Returns the voice locale of the current locale."""
        return self._dict().get(self._locale)

    @property
    def arabic(self) -> Optional[VoiceLine]:
        """:class:`Optional[VoiceLine]`: Returns the Arabic locale."""
        return self.ar_AE

    @property
    def german(self) -> Optional[VoiceLine]:
        """:class:`Optional[VoiceLine]`: Returns the German locale."""
        return self.de_DE

    @property
    def english(self) -> Optional[VoiceLine]:
        """:class:`Optional[VoiceLine]`: Returns the English locale."""
        return self.en_US

    @property
    def american_english(self) -> Optional[VoiceLine]:
        """:class:`Optional[VoiceLine]`: Returns the American English locale."""
        return self.en_US

    @property
    def british_english(self) -> Optional[VoiceLine]:
        """:class:`Optional[VoiceLine]`: Returns the British English locale."""
        return self.en_US

    @property
    def spanish(self) -> Optional[VoiceLine]:
        """:class:`Optional[VoiceLine]`: Returns the Spanish locale."""
        return self.es_ES

    @property
    def spanish_mexican(self) -> Optional[VoiceLine]:
        """:class:`Optional[VoiceLine]`: Returns the Spanish Mexican locale."""
        return self.es_MX

    @property
    def french(self) -> Optional[VoiceLine]:
        """:class:`Optional[VoiceLine]`: Returns the French locale."""
        return self.fr_FR

    @property
    def indonesian(self) -> Optional[VoiceLine]:
        """:class:`Optional[VoiceLine]`: Returns the Indonesian locale."""
        return self.id_ID

    @property
    def italian(self) -> Optional[VoiceLine]:
        """:class:`Optional[VoiceLine]`: Returns the Italian locale."""
        return self.it_IT

    @property
    def japanese(self) -> Optional[VoiceLine]:
        """:class:`Optional[VoiceLine]`: Returns the Japanese locale."""
        return self.ja_JP

    @property
    def korean(self) -> Optional[VoiceLine]:
        """:class:`Optional[VoiceLine]`: Returns the Korean locale."""
        return self.ko_KR

    @property
    def polish(self) -> Optional[VoiceLine]:
        """:class:`Optional[VoiceLine]`: Returns the Polish locale."""
        return self.pl_PL

    @property
    def portuguese_brazil(self) -> Optional[VoiceLine]:
        """:class:`Optional[VoiceLine]`: Returns the Portuguese Brazil locale."""
        return self.pt_BR

    @property
    def russian(self) -> Optional[VoiceLine]:
        """:class:`Optional[VoiceLine]`: Returns the Russian locale."""
        return self.ru_RU

    @property
    def thai(self) -> Optional[VoiceLine]:
        """:class:`Optional[VoiceLine]`: Returns the Thai locale."""
        return self.th_TH

    @property
    def turkish(self) -> Optional[VoiceLine]:
        """:class:`Optional[VoiceLine]`: Returns the Turkish locale."""
        return self.tr_TR

    @property
    def vietnamese(self) -> Optional[VoiceLine]:
        """:class:`Optional[VoiceLine]`: Returns the Vietnamese locale."""
        return self.vi_VN

    @property
    def chinese_simplified(self) -> Optional[VoiceLine]:
        """:class:`Optional[VoiceLine]`: Returns the Chinese Simplified locale."""
        return self.zh_CN

    @property
    def chinese_traditional(self) -> Optional[VoiceLine]:
        """:class:`Optional[VoiceLine]`: Returns the Chinese Traditional locale."""
        return self.zh_TW


class Agent(BaseModel):
    def __init__(self, *, state: CacheState, data: AgentPayload) -> None:
        super().__init__(data['uuid'])
        self._state: CacheState = state
        self._display_name: Union[str, Dict[str, str]] = data['displayName']
        self._description: Union[str, Dict[str, str]] = data['description']
        self.developer_name: str = data['developerName']
        self.character_tags: Optional[List[Union[str, Dict[str, str]]]] = data['characterTags']
        self._display_icon: str = data['displayIcon']
        self._display_icon_small: str = data['displayIconSmall']
        self._bust_portrait: str = data['bustPortrait']
        self._full_portrait: str = data['fullPortrait']
        self._full_portrait_v2: str = data['fullPortraitV2']
        self._killfeed_portrait: str = data['killfeedPortrait']
        self._background: str = data['background']
        self.background_gradient_colors: List[str] = data['backgroundGradientColors']
        self.asset_path: str = data['assetPath']
        self._is_full_portrait_right_facing: bool = data['isFullPortraitRightFacing']
        self._is_playable_character: bool = data['isPlayableCharacter']
        self._is_available_for_test: bool = data['isAvailableForTest']
        self._is_base_content: bool = data['isBaseContent']
        self._roles: Role = Role(state=self._state, data=data['role'])
        self._abilities: Dict[str, Ability] = {
            ability['slot'].lower(): Ability(state=self._state, data=ability, agent=self) for ability in data['abilities']
        }
        self._voice_line: Union[VoiceLinePayload, Dict[str, Optional[VoiceLinePayload]]] = data['voiceLine']
        self._display_name_localized: Localization = Localization(self._display_name, locale=self._state.locale)
        self._description_localized: Localization = Localization(self._description, locale=self._state.locale)

    def __str__(self) -> str:
        return self.display_name.locale

    def __repr__(self) -> str:
        return f'<Agent display_name={self.display_name!r}>'

    def display_name_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        return self._display_name_localized.from_locale(locale)

    def description_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        return self._description_localized.from_locale(locale)

    @property
    def display_name(self) -> Localization:
        """:class: `str` Returns the agent's name."""
        return self._display_name_localized

    @property
    def description(self) -> Localization:
        """:class: `str` Returns the agent's description."""
        return self._description_localized

    @property
    def display_icon(self) -> Asset:
        """:class: `Asset` Returns the agent's display icon."""
        return Asset._from_url(state=self._state, url=self._display_icon)

    @property
    def display_icon_small(self) -> Asset:
        """:class: `Asset` Returns the agent's display icon small."""
        return Asset._from_url(state=self._state, url=self._display_icon_small)

    @property
    def bust_portrait(self) -> Asset:
        """:class: `Asset` Returns the agent's bust portrait."""
        return Asset._from_url(state=self._state, url=self._bust_portrait)

    @property
    def full_portrait(self) -> Asset:
        """:class: `Asset` Returns the agent's full portrait."""
        return Asset._from_url(state=self._state, url=self._full_portrait)

    @property
    def full_portrait_v2(self) -> Asset:
        """:class: `Asset` Returns the agent's full portrait v2."""
        return Asset._from_url(state=self._state, url=self._full_portrait_v2)

    @property
    def killfeed_portrait(self) -> Asset:
        """:class: `Asset` Returns the agent's killfeed portrait."""
        return Asset._from_url(state=self._state, url=self._killfeed_portrait)

    @property
    def background(self) -> Asset:
        """:class: `Asset` Returns the agent's background."""
        return Asset._from_url(state=self._state, url=self._background)

    @property
    def role(self) -> Role:
        """:class: `AgentRole` Returns the agent's role."""
        return self._roles

    @property
    def abilities(self) -> List[Ability]:
        """:class: `List[AgentAbility]` Returns the agent's abilities."""
        return list(self._abilities.values())

    @property
    def voice_line(self) -> Optional[VoiceLine]:
        """:class: `AgentVoiceLineLocalization` Returns the agent's voice line."""
        return self.voice_line_localization.voice_locale

    @property
    def voice_line_localization(self) -> VoiceLineLocalization:
        """:class: `AgentVoiceLineLocalization` Returns the agent's voice line."""
        return VoiceLineLocalization(self._voice_line)

    def get_ability(self, ability_type: AbilityType) -> Optional[Ability]:
        """:class: `AgentAbility` Returns the agent's ability from the slot."""
        return self._abilities.get(ability_type.value.lower())

    def is_full_portrait_right_facing(self) -> bool:
        """:class: `bool` Returns whether the agent's full portrait is right facing."""
        return self._is_full_portrait_right_facing

    def is_playable_character(self) -> bool:
        """:class: `bool` Returns whether the agent is a playable character."""
        return self._is_playable_character

    def is_available_for_test(self) -> bool:
        """:class: `bool` Returns whether the agent is available for test."""
        return self._is_available_for_test

    def is_base_content(self) -> bool:
        """:class: `bool` Returns whether the agent is base content."""
        return self._is_base_content



"""
The MIT License (MIT)

Copyright (c) 2023-present STACiA

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Generic, List, Optional, TypeVar, Union

from ..asset import Asset
from ..enums import Locale
from ..localization import Localization
from .abc import BaseModel

if TYPE_CHECKING:
    from typing_extensions import Self

    from ..cache import CacheState
    from ..types.buddies import Buddy as BuddyPayload, BuddyLevel as BuddyLevelPayload
    from .themes import Theme

__all__ = (
    'Buddy',
    'BuddyLevel',
)

BuddyT = TypeVar('BuddyT', bound='Buddy')


class Buddy(BaseModel):
    def __init__(self, *, state: CacheState, data: BuddyPayload) -> None:
        super().__init__(data['uuid'])
        self._state: CacheState = state
        self._display_name: Union[str, Dict[str, str]] = data['displayName']
        self._is_hidden_if_not_owned: bool = data['isHiddenIfNotOwned']
        self._theme_uuid: Optional[str] = data['themeUuid']
        self._display_icon: Optional[str] = data['displayIcon']
        self.asset_path: str = data['assetPath']
        self.levels: List[BuddyLevel] = [BuddyLevel(state=self._state, data=level, parent=self) for level in data['levels']]
        self._name_localized = Localization(self._display_name, locale=self._state.locale)

    def __str__(self) -> str:
        return self.display_name.locale

    def __repr__(self) -> str:
        return f'<Buddy display_name={self.display_name!r}>'

    def display_name_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        return self._name_localized.from_locale(locale)

    @property
    def display_name(self) -> Localization:
        """:class: `str` Returns the buddy's name."""
        return self._name_localized

    @property
    def theme(self) -> Optional[Theme]:
        """:class: `Theme` Returns the buddy's theme."""
        # Avoid circular import
        return self._state.get_theme(self._theme_uuid)

    @property
    def display_icon(self) -> Asset:
        """:class: `Asset` Returns the buddy's icon."""
        return Asset._from_url(state=self._state, url=self._display_icon)

    def is_hidden_if_not_owned(self) -> bool:
        """:class: `bool` Returns whether the buddy is hidden if not owned."""
        return self._is_hidden_if_not_owned

    # helper methods

    def get_buddy_level(self, level: int = 1) -> Optional[BuddyLevel]:
        """Returns the buddy level for the given level number."""
        return next((b for b in self.levels if b.charm_level == level), None)

    @classmethod
    def _copy(cls, buddy: Self) -> Self:
        """Creates a copy of the given buddy."""
        self = cls.__new__(cls)  # bypass __init__
        self._uuid = buddy._uuid
        self._state = buddy._state
        self._display_name = buddy._display_name
        self._is_hidden_if_not_owned = buddy._is_hidden_if_not_owned
        self._theme_uuid = buddy._theme_uuid
        self._display_icon = buddy._display_icon
        self.asset_path = buddy.asset_path
        self.levels = buddy.levels.copy()
        self._name_localized = buddy._name_localized
        return self


class BuddyLevel(BaseModel, Generic[BuddyT]):
    def __init__(self, *, state: CacheState, data: BuddyLevelPayload, parent: BuddyT) -> None:
        super().__init__(data['uuid'])
        self._state: CacheState = state
        self._data: BuddyLevelPayload = data
        self.charm_level: int = data['charmLevel']
        self._display_name: Union[str, Dict[str, str]] = data['displayName']
        self._display_icon: Optional[str] = data['displayIcon']
        self.asset_path: str = data['assetPath']
        self.parent: BuddyT = parent
        self._display_name_localized: Localization = Localization(self._display_name, locale=self._state.locale)

    def __str__(self) -> str:
        return self.display_name.locale

    def __repr__(self) -> str:
        return f'<BuddyLevel display_name={self.display_name!r}>'

    def display_name_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        return self._display_name_localized.from_locale(locale)

    @property
    def level(self) -> int:
        """:class: `int` alias for :attr: `BuddyLevel.charm_level`"""
        return self.charm_level

    @property
    def display_name(self) -> Localization:
        """:class: `str` Returns the buddy's name."""
        return self._display_name_localized

    @property
    def display_icon(self) -> Asset:
        """:class: `str` Returns the buddy's display icon."""
        return Asset._from_url(state=self._state, url=self._display_icon)

    @classmethod
    def _copy(cls, buddy_level: Self) -> Self:
        """Creates a copy of the given buddy level."""
        self = cls.__new__(cls)  # bypass __init__
        self._uuid = buddy_level._uuid
        self._state = buddy_level._state
        self._data = buddy_level._data.copy()
        self.charm_level = buddy_level.charm_level
        self._display_name = buddy_level._display_name
        self._display_icon = buddy_level._display_icon
        self.asset_path = buddy_level.asset_path
        self.parent = buddy_level.parent._copy(buddy_level.parent)
        self._display_name_localized = buddy_level._display_name_localized
        return self



import pytest

from .conftest import BaseTest


class TestInGameAPI(BaseTest):
    @pytest.mark.asyncio
    async def test_client_init(self) -> None:
        assert self.client is not None
        await self.client.init()

    @pytest.mark.asyncio
    async def test_client(self) -> None:
        assert self.client is not None
        assert self.client.is_ready() is True
        assert self.client.is_closed() is False

        for key in self.client.cache.__dict__.keys():
            if key.startswith('_') and isinstance(self.client.cache.__dict__[key], dict):
                assert len(self.client.cache.__dict__[key]) > 0

        await self.client.close()
        assert self.client.is_closed() is True
        self.client.clear()

        for key in self.client.cache.__dict__.keys():
            if key.startswith('_') and isinstance(self.client.cache.__dict__[key], dict):
                assert len(self.client.cache.__dict__[key]) == 0

        assert self.client.is_ready() is False
        assert self.client.is_closed() is False

    @pytest.mark.asyncio
    async def test_close(self) -> None:
        await self.client.close()



"""
The MIT License (MIT)

Copyright (c) 2023-present STACiA

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Optional, Union

from .. import utils
from ..asset import Asset
from ..enums import Locale
from ..localization import Localization
from .abc import BaseModel

if TYPE_CHECKING:
    from ..cache import CacheState
    from ..types.gamemodes import (
        GameFeatureOverride as GameFeatureOverridePayload,
        GameMode as GameModePayload,
        GameModeEquippable as GameModeEquippablePayload,
        GameRuleBoolOverride as GameRuleBoolOverridePayload,
    )
    from .weapons import Weapon

# fmt: off
__all__ = (
    'GameFeatureOverride',
    'GameMode',
    'GameModeEquippable',
    'GameRuleBoolOverride',
)
# fmt: on


class GameFeatureOverride:
    def __init__(self, data: GameFeatureOverridePayload) -> None:
        self.feature_name: str = data['featureName']
        self.state: bool = data['state']

    def __repr__(self) -> str:
        return f'<GameFeatureOverride feature_name={self.feature_name!r} state={self.state!r}>'

    def __bool__(self) -> bool:
        return self.state


class GameRuleBoolOverride:
    def __init__(self, data: GameRuleBoolOverridePayload) -> None:
        self.rule_name: str = data['ruleName']
        self.state: bool = data['state']

    def __bool__(self) -> bool:
        return self.state

    def __repr__(self) -> str:
        return f'<GameRuleBoolOverride rule_name={self.rule_name!r} state={self.state!r}>'


class GameMode(BaseModel):
    def __init__(self, state: CacheState, data: GameModePayload) -> None:
        super().__init__(data['uuid'])
        self._state: CacheState = state
        self._display_name: Union[str, Dict[str, str]] = data['displayName']
        self._duration: Union[str, Dict[str, str]] = data['duration']
        self.allows_match_timeouts: bool = data['allowsMatchTimeouts']
        self._is_team_voice_allowed: bool = data['isTeamVoiceAllowed']
        self._is_minimap_hidden: bool = data['isMinimapHidden']
        self.orb_count: int = data['orbCount']
        self.rounds_per_half: int = data['roundsPerHalf']
        self.team_roles: Optional[List[str]] = data['teamRoles']
        self.game_feature_overrides: Optional[List[GameFeatureOverride]] = None
        if data['gameFeatureOverrides'] is not None:
            self.game_feature_overrides = [GameFeatureOverride(feature) for feature in data['gameFeatureOverrides']]
        self.game_rule_bool_overrides: Optional[List[GameRuleBoolOverride]] = None
        if data['gameRuleBoolOverrides'] is not None:
            self.game_rule_bool_overrides = [GameRuleBoolOverride(rule) for rule in data['gameRuleBoolOverrides']]
        self._display_icon: Optional[str] = data['displayIcon']
        self.asset_path: str = data['assetPath']
        self._display_name_localized: Localization = Localization(self._display_name, locale=self._state.locale)
        self._duration_localized: Localization = Localization(self._duration, locale=self._state.locale)

    def __str__(self) -> str:
        return self.display_name.locale

    def __repr__(self) -> str:
        return f'<GameMode display_name={self.display_name!r}>'

    def __eq__(self, other: object) -> bool:
        return isinstance(other, GameMode) and self.uuid == other.uuid

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    # def __contains__(self, item: Union[GameMode, GameModeType]) -> bool:
    #     ...

    def display_name_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        return self._display_name_localized.from_locale(locale)

    def duration_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        return self._duration_localized.from_locale(locale)

    @property
    def display_name(self) -> Localization:
        """:class: `str` Returns the game mode's name."""
        return self._display_name_localized

    @property
    def duration(self) -> Localization:
        """:class: `str` Returns the game mode's duration."""
        return self._duration_localized

    @property
    def display_icon(self) -> Optional[Asset]:
        """:class: `Asset` Returns the game mode's display icon."""
        if self._display_icon is None:
            return None
        return Asset(self._state, self._display_icon)

    def is_team_voice_allowed(self) -> bool:
        """:class: `bool` Returns whether the game mode allows team voice."""
        return self._is_team_voice_allowed

    def is_minimap_hidden(self) -> bool:
        """:class: `bool` Returns whether the game mode hides the minimap."""
        return self._is_minimap_hidden


class GameModeEquippable(BaseModel):
    def __init__(self, state: CacheState, data: GameModeEquippablePayload) -> None:
        super().__init__(data['uuid'])
        self._state: CacheState = state
        self._display_name: Union[str, Dict[str, str]] = data['displayName']
        self._category: Optional[str] = data['category']
        self._display_icon: str = data['displayIcon']
        self._kill_stream_icon: str = data['killStreamIcon']
        self.asset_path: str = data['assetPath']
        self._display_name_localized: Localization = Localization(self._display_name, locale=self._state.locale)

    def __str__(self) -> str:
        return self.display_name.locale

    def __repr__(self) -> str:
        return f'<GameModeEquippable display_name={self.display_name!r}>'

    def display_name_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        return self._display_name_localized.from_locale(locale)

    @property
    def display_name(self) -> Localization:
        """:class: `str` Returns the game mode's name."""
        return self._display_name_localized

    @property
    def category(self) -> Optional[str]:
        """:class: `str` Returns the game mode's category."""
        return utils.removeprefix(self._category, 'EEquippableCategory::') if self._category else None

    @property
    def display_icon(self) -> Asset:
        """:class: `Asset` Returns the game mode's display icon."""
        return Asset._from_url(self._state, url=self._display_icon)

    @property
    def kill_stream_icon(self) -> Asset:
        """:class: `Asset` Returns the game mode's kill stream icon."""
        return Asset._from_url(self._state, url=self._kill_stream_icon)

    def weapon(self) -> Optional[Weapon]:
        """:class: `Weapon` Returns the game mode's weapon."""
        # special weapon for game mode
        return self._state.get_weapon(self._uuid)



"""
Valorant API Wrapper
~~~~~~~~~~~~~~~~~~~
A basic wrapper for the Valorant API.
:copyright: (c) 2023-present STACiA
:license: MIT, see LICENSE for more details.
"""

__title__ = 'valorant'
__author__ = 'STACiA'
__license__ = 'MIT'
__copyright__ = 'Copyright 2023-present STACiA'
__version__ = '1.0.1'

from typing import Literal, NamedTuple

from . import models as models, utils as utils
from .asset import *
from .client import *
from .enums import *
from .localization import *
from .models import *


class VersionInfo(NamedTuple):
    major: int
    minor: int
    micro: int
    release: Literal["alpha", "beta", "final"]


version_info: VersionInfo = VersionInfo(major=1, minor=0, micro=1, release='final')

del NamedTuple, Literal, VersionInfo



"""
The MIT License (MIT)

Copyright (c) 2023-present STACiA

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Dict, List, Optional, Union

from .. import utils
from ..enums import Locale, MissionType, try_enum
from ..localization import Localization
from .abc import BaseModel

if TYPE_CHECKING:
    from ..cache import CacheState
    from ..types.missions import Mission as MissionPayload, Objective as ObjectivePayload

# fmt: off
__all__ = (
    'Mission',
)
# fmt: on


class Objective(BaseModel):
    def __init__(self, data: ObjectivePayload) -> None:
        super().__init__(data['objectiveUuid'])
        self.value: int = data['value']

    # TODO: add magic methods


class Mission(BaseModel):
    def __init__(self, state: CacheState, data: MissionPayload) -> None:
        super().__init__(data['uuid'])
        self._state: CacheState = state
        self._data: MissionPayload = data
        self._display_name: Optional[Union[str, Dict[str, str]]] = data['displayName']
        self._title: Optional[Union[str, Dict[str, str]]] = data['title']
        self._type: Optional[str] = data['type']
        self.xp_grant: int = data['xpGrant']
        self.progress_to_complete: int = data['progressToComplete']
        self._activation_date_iso: str = data['activationDate']
        self._expiration_date_iso: str = data['expirationDate']
        self.tags: Optional[List[str]] = data['tags']
        self.objectives: Optional[List[Objective]] = None
        if data['objectives'] is not None:
            self.objectives = [Objective(obj) for obj in data['objectives']]
        self.asset_path: str = data['assetPath']
        self._display_name_localized: Localization = Localization(self._display_name, locale=self._state.locale)
        self._title_localized: Localization = Localization(self._title, locale=self._state.locale)

    def __str__(self) -> str:
        return self.title.locale

    def __repr__(self) -> str:
        return f'<Mission title={self.title!r}>'

    def __int__(self) -> int:
        return self.xp_grant

    @property
    def xp(self) -> int:
        """:class: `int` alias for :attr: `Mission.xp_grant`"""
        return self.xp_grant

    def display_name_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        return self._display_name_localized.from_locale(locale)

    def title_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        return self._title_localized.from_locale(locale)

    @property
    def display_name(self) -> Localization:
        """:class: `str` Returns the mission's name."""
        return self._display_name_localized

    @property
    def title(self) -> Localization:
        """:class: `str` Returns the mission's title."""
        return self._title_localized

    @property
    def type(self) -> Optional[MissionType]:
        """Optional[:class: `MissionType`] Returns the mission's type."""
        if self._type == '' or self._type is None:
            return None
        return try_enum(MissionType, utils.removeprefix(self._type, 'EAresMissionType::'))

    @property
    def activation_date(self) -> Optional[datetime.datetime]:
        """:class: `datetime.datetime` Returns the mission's activation date."""
        return utils.parse_iso_datetime(self._activation_date_iso) if self._activation_date_iso else None

    @property
    def expiration_date(self) -> Optional[datetime.datetime]:
        """:class: `datetime.datetime` Returns the mission's expiration date."""
        return utils.parse_iso_datetime(self._expiration_date_iso) if self._expiration_date_iso else None



import re

from setuptools import setup

# inspired by https://github.com/Rapptz/discord.py/blob/master/setup.py

requirements = []
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

version = ''
with open('valorant/__init__.py') as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)  # type: ignore

readme = ''
with open('README.md') as f:
    readme = f.read()

extras_require = {
    'speed': ['orjson>=3.8.11,<4.0'],
}

packages = [
    'valorant',
    'valorant.models',
    'valorant.types',
]

setup(
    name='valorant.py',
    author='STACiA',
    url='https://github.com/staciax/valorant',
    project_urls={
        "Issue tracker": "https://github.com/staciax/valorant/issues",
    },
    version=version,
    packages=packages,
    license='MIT',
    description='An Asynchronous Unofficial Valorant API Wrapper for Python',
    long_description=readme,
    long_description_content_type='text/markdown',
    include_package_data=True,
    install_requires=requirements,
    extras_require=extras_require,
    python_requires='>=3.8.0',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
)



"""
The MIT License (MIT)

Copyright (c) 2023-present STACiA

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Optional, Union

from ..asset import Asset
from ..enums import Locale
from ..localization import Localization
from .abc import BaseModel

if TYPE_CHECKING:
    from ..cache import CacheState
    from ..types.themes import Theme as ThemePayload

# fmt: off
__all__ = (
    'Theme',
)
# fmt: on


class Theme(BaseModel):
    def __init__(self, state: CacheState, data: ThemePayload) -> None:
        super().__init__(data['uuid'])
        self._state: CacheState = state
        self._display_name: Union[str, Dict[str, str]] = data['displayName']
        self._display_icon: Optional[str] = data['displayIcon']
        self._store_featured_image: Optional[str] = data['storeFeaturedImage']
        self.asset_path: str = data['assetPath']
        self._display_name_localized: Localization = Localization(self._display_name, locale=self._state.locale)

    def __str__(self) -> str:
        return self.display_name.locale

    def __repr__(self) -> str:
        return f'<Theme display_name={self.display_name!r}>'

    def display_name_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        return self._display_name_localized.from_locale(locale)

    @property
    def display_name(self) -> Localization:
        """:class: `str` Returns the ceremony's name."""
        return self._display_name_localized

    @property
    def display_icon(self) -> Optional[Asset]:
        """:class: `Asset` Returns the ceremony's display icon."""
        if self._display_icon is None:
            return None
        return Asset._from_url(self._state, self._display_icon)

    @property
    def store_featured_image(self) -> Optional[Asset]:
        """:class: `Asset` Returns the ceremony's store featured image."""
        if self._store_featured_image is None:
            return None
        return Asset._from_url(self._state, self._store_featured_image)



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


# source : https://github.com/Rapptz/discord.py/blob/master/discord/asset.py
from __future__ import annotations

import io
import os
from typing import TYPE_CHECKING, Any, Optional, Tuple, Union

import yarl

from . import utils
from .file import File

if TYPE_CHECKING:
    from typing_extensions import Self

    from .cache import CacheState

MISSING = utils.MISSING


class AssetMixin:
    __slots__ = ()
    url: str
    _state: CacheState

    async def read(self):
        """|coro|
        Retrieves the content of this asset as a :class:`bytes` object.
        Raises
        ------
        Exception
            There was no internal connection state.
        HTTPException
            Downloading the asset failed.
        NotFound
            The asset was deleted.
        Returns
        -------
        :class:`bytes`
            The content of the asset.
        """
        if self._state is None:
            raise ValueError('Asset has no client')

        return await self._state.http.read_from_url(self.url)

    async def save(self, fp: Union[str, bytes, os.PathLike[Any], io.BufferedIOBase], *, seek_begin: bool = True) -> int:
        """|coro|
        Saves this asset into a file-like object.
        Parameters
        ----------
        fp: Union[:class:`io.BufferedIOBase`, :class:`os.PathLike`]
            The file-like object to save this asset to or the filename
            to use. If a filename is passed then a file is created with that
            filename and used instead.
        seek_begin: :class:`bool`
            Whether to seek to the beginning of the file after saving is
            successfully done.
        Raises
        ------
        Exception
            There was no internal connection state.
        HTTPException
            Downloading the asset failed.
        NotFound
            The asset was deleted.
        Returns
        --------
        :class:`int`
            The number of bytes written.
        """
        data = await self.read()
        if isinstance(fp, io.BufferedIOBase):
            written = fp.write(data)
            if seek_begin:
                fp.seek(0)
            return written
        else:
            with open(fp, 'wb') as f:
                return f.write(data)

    async def to_file(self, *, filename: Optional[str] = MISSING) -> File:
        """|coro|
        Converts the asset into a :class:`File`
        Parameters
        -----------
        filename: Optional[:class:`str`]
            The filename of the file. If not provided, then the filename from
            the asset's URL is used.
        Raises
        ------
        Exception
            The asset does not have an associated state.
        ValueError
            The asset is a unicode emoji.
        TypeError
            The asset is a sticker with lottie type.
        HTTPException
            Downloading the asset failed.
        NotFound
            The asset was deleted.
        Returns
        -------
        :class:`File`
            The asset as a file suitable for sending.
        """
        data = await self.read()
        file_filename = filename if filename is not MISSING else yarl.URL(self.url).name
        return File(io.BytesIO(data), filename=file_filename)


class Asset(AssetMixin):

    """Represents a CDN asset on Valorant.
    .. container:: operations
        .. describe:: str(x)
            Returns the URL of the CDN asset.
        .. describe:: len(x)
            Returns the length of the CDN asset's URL.
        .. describe:: x == y
            Checks if the asset is equal to another asset.
        .. describe:: x != y
            Checks if the asset is not equal to another asset.
        .. describe:: hash(x)
            Returns the hash of the asset.
    """

    __slot__: Tuple[str, ...] = ('_client', '_url', '_video', '_animated')

    def __init__(self, state: CacheState, url: str, *, animated: bool = False, video: bool = False) -> None:
        self._state = state
        self._url = url
        self._video = video
        self._animated = animated

    @classmethod
    def _from_url(cls, state: CacheState, url: Optional[str] = None, *, animated: bool = False) -> Self:
        if url is None:
            raise TypeError('Expected URL, not NoneType')

        video = True if url.endswith('.mp4') else False
        if not animated and url.endswith('.gif'):
            animated = True
        return cls(state=state, url=url, animated=animated, video=video)

    def __str__(self) -> str:
        return str(self._url)

    def __len__(self) -> int:
        return len(self._url)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Asset) and self._url == other._url

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(self._url)

    @property
    def url(self) -> str:
        """:class:`str`: Returns the underlying URL of the asset."""
        return self._url

    @property
    def animated(self) -> bool:
        """:class:`bool`: Returns whether the asset is animated."""
        return self._animated

    @property
    def video(self) -> bool:
        """:class:`bool`: Returns whether the asset is a video."""
        return self._video



"""
The MIT License (MIT)

Copyright (c) 2023-present STACiA

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Optional, Union

from ..enums import Locale
from ..localization import Localization
from .abc import BaseModel

if TYPE_CHECKING:
    from typing_extensions import Self

    from ..cache import CacheState
    from ..types.player_titles import PlayerTitle as PlayerTitlePayload


# fmt: off
__all__ = (
    'PlayerTitle',
)
# fmt: on


class PlayerTitle(BaseModel):
    def __init__(self, *, state: CacheState, data: PlayerTitlePayload) -> None:
        super().__init__(data['uuid'])
        self._state: CacheState = state
        self._data: PlayerTitlePayload = data
        self._display_name: Union[str, Dict[str, str]] = data['displayName']
        self._title_text: Union[str, Dict[str, str]] = data['titleText']
        self._is_hidden_if_not_owned: bool = data.get('isHiddenIfNotOwned', False)
        self.asset_path: str = data['assetPath']
        self._display_name_localized: Localization = Localization(self._display_name, locale=self._state.locale)
        self._title_text_localized: Localization = Localization(self._title_text, locale=self._state.locale)

    def __str__(self) -> str:
        return self.text.locale

    def __repr__(self) -> str:
        return f"<PlayerTitle display_name={self.display_name!r} text={self.text!r}>"

    def display_name_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        return self._display_name_localized.from_locale(locale)

    def title_text_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        return self._title_text_localized.from_locale(locale)

    @property
    def display_name(self) -> Localization:
        """:class: `str` Returns the player title's name."""
        return self._display_name_localized

    @property
    def text(self) -> Localization:
        """:class: `str` Returns the player title's title text."""
        return self._title_text_localized

    def is_hidden_if_not_owned(self) -> bool:
        """:class: `bool` Returns whether the player title is hidden if not owned."""
        return self._is_hidden_if_not_owned

    @classmethod
    def _copy(cls, player_title: Self) -> Self:
        self = cls.__new__(cls)  # bypass __init__
        self._uuid = player_title._uuid
        self._state = player_title._state
        self._data = player_title._data
        self._display_name = player_title._display_name
        self._title_text = player_title._title_text
        self._is_hidden_if_not_owned = player_title._is_hidden_if_not_owned
        self.asset_path = player_title.asset_path
        self._display_name_localized = player_title._display_name_localized
        self._title_text_localized = player_title._title_text_localized
        return self
